//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// element.h

//------------------------------------------------------------------------
// global variables - those which do not depend on class definitions
//------------------------------------------------------------------------

extern char variables[];         // names of variables
extern int integers[];           // in denominators
extern bool argument[];          // true if this function depends on this coord
extern int function_power[];     // f^function_power[i] = *function_terms[i]
extern int number_of_passes;     // number of simplification passes
extern int number_to_simplify;   // number of functions to simplify this pass
extern int pass_num_fns[];       // number of functions in each pass
extern int pass_fns[];           // specific functions in all passes
extern int pass_fn_index;        // index into pass_fns for this pass
extern int sort_variable[];      // leading variable in each function

//------------------------------------------------------------------------
// class element - elements in a doubly-linked list
//------------------------------------------------------------------------

class element {

 protected:

  element* next;
  element* previous;

 public:

#ifdef debug
  int element_magic_no;
#endif

  inline void element_init() {
#ifdef tracentry
    trace_ei += 1;
#endif
#ifdef debug
    element_magic_no = element_magic;
#endif
    next = NULL;
    previous = NULL;
  }

  inline element() {
#ifdef tracentry
    trace_e += 1;
#endif
    element_init();
  }

  inline ~element() {
#ifdef tracentry
    trace_ke += 1;
#endif
#ifdef debug
    if ((next != NULL) || (previous != NULL))
      die("attempting to delete linked element");
    element_magic_no = -element_magic;
#endif
  }

  inline element* next_element() {
#ifdef tracentry
    trace_ne += 1;
#endif
#ifdef debug
    if ((next != NULL) && (next->element_magic_no != element_magic))
      die("next element is not an element");
#endif
    return next; }

  inline void insert_element(element* before) {    // assume insertion at list head
#ifdef tracentry
    trace_ie += 1;
#endif
#ifdef debug
    if (before != NULL) {
      if (before->element_magic_no != element_magic)
	die("attempting to insert before a non-element");
      if (before->previous != NULL)
	die("attempting to insert in middle of list");}
#endif
    next = before;
    if (before != NULL) {
#ifdef debug
      if ((before->previous != NULL) && ((before->previous)->element_magic_no != element_magic))
	die("previous does not link to element in insert");
#endif
      before->previous = this;}
    previous = NULL; }
  
  inline void remove_element() {
#ifdef tracentry
    trace_re += 1;
#endif
    if (previous != NULL) {
#ifdef debug
      if (previous->element_magic_no != element_magic)
	die("previous does not link to element in remove");
#endif
      previous->next = next;}
    if (next != NULL) {
#ifdef debug
      if (next->element_magic_no != element_magic)
	die("next does not link to element in remove");
#endif
      next->previous = previous;}
#ifdef debug
    next = NULL;
    previous = NULL;
#endif
    }
};

//------------------------------------------------------------------------
// class term - element describing a term in a sum
//------------------------------------------------------------------------

class term : public element {

 public:

#ifdef debug
  int term_magic_no;
#endif

  long long coefficient;
  int* factor_powers;         // one integer power for each variable (or function)

  void send ();
  
  inline term() {             // must be here to satisfy c++
#ifdef tracentry
      trace_td += 1;
#endif
#ifdef debug
    term_magic_no = term_magic;
#endif
    element_init();
    factor_powers = NULL; }

  inline term(long long *coefficient_param, int* factor_powers_param) {
#ifdef tracentry
    trace_t += 1;
#endif
#ifdef debug
    term_magic_no = term_magic;
#endif
    element_init();
    coefficient = *coefficient_param;
    factor_powers = new int[number_of_variables];
    if (factor_powers == NULL)
      die("allocating factor_powers");
    memcpy(factor_powers, factor_powers_param, number_of_variables * sizeof(int)); }
  
  inline ~term() {
#ifdef tracentry
    trace_kt += 1;
#endif
#ifdef debug
    term_magic_no = -term_magic;
#endif
    delete [] factor_powers;
    factor_powers = NULL; }
  
  inline term* next_term() {
#ifdef tracentry
    trace_nt += 1;
#endif
    return (term*) next_element(); }

  inline term* insert_term(term* before) {
#ifdef tracentry
    trace_it += 1;
#endif
#ifdef debug
    if (before->term_magic_no != term_magic)
      die("before is not a term in insert");
#endif
    insert_element((element*) before);
    return this; }

  inline void fraction_cd(long long* coefficient_param, int *factor_powers_param) {
#ifdef tracentry
    trace_ff += 1;
#endif
    for (int i = number_of_variables - number_of_fractions; i < number_of_variables; i++) {
#ifdef debug
      if (factor_powers[i] > 0)
	die("bad factor_powers in fraction_cd");
      if (factor_powers_param[i] > 0)
	die("bad factor_powers_param in fraction_cd");
#endif
      for (int j = factor_powers[i] - factor_powers_param[i]; j < 0; j++) {
	factor_powers_param[i] -= 1;
	*coefficient_param *= integers[i - (number_of_variables - number_of_fractions)];}
      for (int j = factor_powers_param[i] - factor_powers[i]; j < 0; j++) {
	factor_powers[i] -= 1;
	coefficient *= integers[i - (number_of_variables - number_of_fractions)];}} }

  inline void normalize() {
    long long divided;
#ifdef tracentry
    trace_tf += 1;
#endif
    for (int i = number_of_variables - number_of_fractions; i < number_of_variables; i++)
      for (int j = factor_powers[i]; j < 0; j++) {
	divided = coefficient / integers[i - (number_of_variables - number_of_fractions)];
	if ((integers[i - (number_of_variables - number_of_fractions)] * divided) == coefficient) {
	  coefficient = divided;
	  factor_powers[i] += 1;}
	else
	  break;} }
};

//------------------------------------------------------------------------
// class invariant_term - invariants are also placed on lists by function power
//------------------------------------------------------------------------

class invariant_term : public term {

 protected:

  invariant_term** next_invariant;
  invariant_term** previous_invariant;
  
 public:

#ifdef debug
  int invariant_term_magic_no;
#endif

  inline invariant_term(long long *coeff, int* factor_powers_param) {
#ifdef tracentry
    trace_int += 1;
#endif
#ifdef debug
    invariant_term_magic_no = invariant_term_magic;
#endif
    coefficient = *coeff;
    factor_powers = new int[number_of_variables];
    if (factor_powers == NULL)
      die("allocating factor_powers");
    memcpy(factor_powers, factor_powers_param, number_of_variables * sizeof(int));
    next_invariant = new invariant_term* [number_to_simplify];
    if (next_invariant == NULL)
      die("allocating next_invariant");
    previous_invariant = new invariant_term* [number_to_simplify];
    if (previous_invariant == NULL)
      die("allocating previous invariant");
    for (int i = 0; i < number_to_simplify; i++)
      {next_invariant[i] = NULL;
      previous_invariant[i] = NULL;} }
  
  inline ~invariant_term() {
#ifdef tracentry
    trace_kint += 1;
#endif
#ifdef debug
    for (int i = 0; i < number_to_simplify; i++)
      if ((next_invariant[i] != NULL) || (previous_invariant[i] != NULL))
	die("attempting to delete linked invariant term");
    invariant_term_magic_no = -invariant_term_magic;
#endif
    delete [] next_invariant;
    delete [] previous_invariant;
    next_invariant = NULL;
    previous_invariant = NULL; }
  
  inline invariant_term* next_invariant_term(int function) {
#ifdef tracentry
    trace_nint += 1;
#endif
#ifdef debug
    if ((function < 0) || (function >= number_to_simplify))
      die("next function index out of bounds");
    if (next_invariant[function] != NULL)
      if (next_invariant[function]->invariant_term_magic_no != invariant_term_magic)
	die("next_invariant does not point to an invariant term");
#endif
    return next_invariant[function]; }

  inline void insert_invariant_term(int function, invariant_term* before) {
#ifdef tracentry
    trace_iit += 1;
#endif
#ifdef debug
    if (before != NULL)
      if (before->invariant_term_magic_no != invariant_term_magic)
	die("before is not an invariant term");
    if ((function < 0) || (function >= number_to_simplify))
      die("insert function index out of bounds");
#endif
    next_invariant[function] = before;
    if (before != NULL) {
#ifdef debug
      if (before->previous_invariant[function] != NULL)
	die("attempting to insert invariant in middle of list");
#endif
      before->previous_invariant[function] = this;}
    previous_invariant[function] = NULL;}

  inline void remove_invariant_links() {
#ifdef tracentry
    trace_rinl += 1;
#endif
#ifdef debug
    if (invariant_term_magic_no != invariant_term_magic)
      die("this is not an invariant term");
#endif
    for (int i = 0; i < number_to_simplify; i++) {
      if (previous_invariant[i] != NULL) {            // it's not on the front of the list
#ifdef debug
	if ((previous_invariant[i])->invariant_term_magic_no != invariant_term_magic)
	  die("previous_invariant does not point to an invariant term (2)");
#endif
	(previous_invariant[i])->next_invariant[i] = next_invariant[i];}
      if (next_invariant[i] != NULL) {
#ifdef debug
	if ((next_invariant[i])->invariant_term_magic_no != invariant_term_magic)
	  die("next_invariant does not point to an invariant term (2)");
#endif
	(next_invariant[i])->previous_invariant[i] = previous_invariant[i];}
#ifdef debug
      next_invariant[i] = NULL;
      previous_invariant[i] = NULL;
#endif
      } }
};

void function_build(int numf, int *numtrm, int *numfct, int *fpwr, long long *co);
