//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// hash.h

//------------------------------------------------------------------------
// class hash - array of linked lists which are hashed into by variable powers
//------------------------------------------------------------------------

class hash {

 protected:

  int last_hash;         // last_hash contains the most recent hash code
  element** head;

#ifdef keepstats
#ifdef debug
  int* headcount;
#endif
#endif

 public:

  int size, number_of_components;

#ifdef debug
  int hash_magic_no;
#endif

  int rank;

  inline void hash_code (int* factor_powers_param) {
#ifdef tracentry
      trace_hc += 1;
#endif
    long product = size;
    for (int i = 0; i < rank; i++) {
/*        last_hash = (factor_powers_param[i] + simplify_size + i) * (i + 1); */
      last_hash = factor_powers_param[i] * (i + 1) * ((int) variables[i]);
      product = (product | last_hash) - (product & last_hash);}
/*        product *= factor_powers_param[i] + simplify_size + i; */
/*        product &= 0xffffff;} */
    last_hash = labs (product) % size; }

  inline void set_hash (int i) {
#ifdef tracentry
      trace_sh += 1;
#endif
    last_hash = i; }
  
  inline void insert_at_head (element* element_param) {
#ifdef tracentry
    trace_iah += 1;
#endif
#ifdef debug
    if (element_param->element_magic_no != element_magic)
      die ("attempting to insert a non-element");
    if ((last_hash < 0) || (last_hash >= size))
      die ("insert - last_hash out of bounds");
#endif
    element* this_head = head[last_hash];
    element_param->insert_element (this_head);
#ifdef keepstats
#ifdef debug
    headcount[last_hash] += 1;
#endif
#endif
    head[last_hash] = element_param; }
  
  inline void remove_from_list (element* element_param) {
#ifdef tracentry
    trace_rfl += 1;
#endif
#ifdef debug
    if (element_param->element_magic_no != element_magic)
      die ("attempting to remove a non-element");
    if ((last_hash < 0) || (last_hash >= size))
      die ("remove - last_hash out of bounds");
#endif
    if (head[last_hash] == element_param)        // it's on the front of the list
      head[last_hash] = element_param->next_element ();
#ifdef keepstats
#ifdef debug
    headcount[last_hash] -= 1;
#endif
#endif
    element_param->remove_element (); }
};

//------------------------------------------------------------------------
// class sum - hash object whose elements are terms in a sum
//------------------------------------------------------------------------

class sum : public hash {

 public:

#ifdef debug
  int sum_magic_no;
#endif

  sum ();
  sum (int);
  ~sum ();
  void sum_build (int, int*, int*, long long*);
  void zero ();                                            // zeroes a sum
  void multiply (int, sum**, long long*);                    // adds product of sums to this sum
  void derivative (sum*, int, long long*);                   // adds derivative of a sum to this sum
  void simplify ();                                        // attempts synthetic division to simplify functions
  void simplify_invariant ();
  void display (char*);                                    // sum output
  void display_add (char*);                                // sum output with +=
  void dump (char*);                                       // non-mathematica sum output

  inline void insert_term (term* term_param) {
#ifdef tracentry
    trace_sit += 1;
#endif
#ifdef debug
    if (term_param->term_magic_no != term_magic)
      die ("attempting to insert a non-term");
#endif
    number_of_components += 1;
    insert_at_head ((element*) term_param); }
  
  inline void remove_term (term* term_param) {
#ifdef tracentry
    trace_srt += 1;
#endif
#ifdef debug
    if (term_param->term_magic_no != term_magic)
      die ("attempting to remove a non-term");
#endif
    remove_from_list ((element*) term_param);
    number_of_components -= 1;
    delete term_param;}
  
  inline term* first_term (int i) {
#ifdef tracentry
    trace_ft += 1;
#endif
#ifdef debug
    if ((head[i] != NULL) && (((term*) head[i])->term_magic_no != term_magic))
      die ("first term is not a term");
#endif
    return (term*) head[i]; }

  inline term* find (int* factor_powers_param) {
#ifdef tracentry
    trace_f += 1;
#endif
    term* current_term;
    hash_code (factor_powers_param); 
    current_term = first_term (last_hash);
    while (current_term != NULL) {
#ifdef debug
      if (current_term->term_magic_no != term_magic)
	die ("found a non-term");
#endif
      if (memcmp (current_term->factor_powers, factor_powers_param, 
		 (number_of_variables - number_of_fractions) * sizeof (int)) == 0)
	return current_term;
      current_term = current_term->next_term ();}
    return NULL;}
  
  inline void add (long long* coefficient_param, int *factor_powers_param) {
    long long zero = 0;
#ifdef tracentry
    trace_sa += 1;
#endif
    term* current_term = find (factor_powers_param);
    if (current_term != NULL) {
      current_term->fraction_cd (coefficient_param, factor_powers_param);
      current_term->coefficient += *coefficient_param;
      if (current_term->coefficient == zero) {
	remove_term (current_term);
	return;}
      current_term->normalize ();
      return;}
    current_term = new term (coefficient_param, factor_powers_param);
    if (current_term == NULL)
      die ("allocating term");
    insert_term (current_term); }

  inline void scalar_multiply (long long factor) {
#ifdef tracentry
    trace_smu += 1;
#endif
  term *current_term;
  if (factor == 1)
    return;
  for (int k = 0; k < size; k++) {
    current_term = first_term (k);
    while (current_term != NULL) {
      current_term->coefficient *= factor;
      current_term->normalize ();
      current_term = current_term->next_term ();}} }

  inline void scalar_multiply (int *fraction_powers) {
#ifdef tracentry
    trace_smu += 1;
#endif
#ifdef debug
    for (int i = 0; i < number_of_fractions; i++)
      if (fraction_powers[i] > 0)
	die ("bad param in scalar_multiply by fraction");
#endif
  term *current_term;
  for (int k = 0; k < size; k++) {
    current_term = first_term (k);
    while (current_term != NULL) {
      for (int i = number_of_variables - number_of_fractions; i < number_of_variables; i++)
	current_term->factor_powers[i] += fraction_powers[i - (number_of_variables - number_of_fractions)];
      current_term->normalize ();
      current_term = current_term->next_term ();}} }

  inline bool zeroq () {
#ifdef tracentry
    trace_szq += 1;
#endif
#ifdef debug
    if (sum_magic_no != sum_magic)
      die ("attempting to check a non-sum");
#endif
    for (int i = 0; i < size; i++)
      if (head[i] != NULL) {
#ifdef debug
	if (number_of_components == 0)
	  die ("number of components = 0 in non-empty sum");
#endif
	return false;}
#ifdef debug
    if (number_of_components > 0)
      die ("nonzero number of components in empty sum");
#endif
    return true; }
      
};

//------------------------------------------------------------------------
// class invariant - sum which represents an invariant
//------------------------------------------------------------------------

class invariant : public sum {

 protected:

  invariant_term** invariant_head;

 public:
 
#ifdef debug
  int invariant_magic_no;
#endif

  invariant ();
  ~invariant ();
  void zero ();
  void multiply (int, sum**, long long*);

  inline invariant_term* first_invariant (int fn, int pow) {
#ifdef tracentry
    trace_fit += 1;
#endif
#ifdef debug
    if ((fn < 0) || (fn >= number_to_simplify))
      die ("bad function parameter in first_invariant");
    if ((pow < 0) || (pow >= simplify_size))
      die ("bad power parameter in first_invariant");
#endif
    return invariant_head[fn * simplify_size + pow]; }
  
  inline void remove_term (invariant_term* term_param) {
#ifdef tracentry
    trace_irt += 1;
#endif
#ifdef debug
    if (term_param->invariant_term_magic_no != invariant_term_magic)
      die ("attempting to remove a non-invariant_term");
#endif
    remove_from_list ((element*) term_param);
    number_of_components -= 1;
    delete term_param;}

  inline void add (long long* coefficient_param, int *factor_powers_param) {
    long long zero = 0;
#ifdef tracentry
    trace_ia += 1;
#endif
    int power, svi;
    invariant_term* current_term = (invariant_term*) find (factor_powers_param);
    if (current_term != NULL) {
      current_term->fraction_cd (coefficient_param, factor_powers_param);
      current_term->coefficient += *coefficient_param;
      if (current_term->coefficient == zero) {
	remove_from_invariant (current_term);
	remove_term (current_term);
	return;}
      current_term->normalize ();
      return;}
    current_term = new invariant_term (coefficient_param, factor_powers_param);
    if (current_term == NULL)
      die ("allocating invariant term");
    insert_term ((term *) current_term);
    for (int i = 0; i < number_to_simplify; i++) {
      svi = pass_fns[i + pass_fn_index];
#ifdef debug
      if ((svi < 0) || (svi > number_of_functions))
	die ("bad sort variable index in inv:add");
#endif
      if (sort_variable[svi] >= 0) {
	power = abs (current_term->factor_powers[sort_variable[svi]]);
	if (power >= simplify_size)
	  die ("add - invariant power exceeds simplify_size");
	current_term->insert_invariant_term (i,invariant_head[i * simplify_size + power]);
#ifdef debug
	if (current_term->invariant_term_magic_no != invariant_term_magic)
	  die ("invariant became a non-invariant");
#endif
	invariant_head[i * simplify_size + power] = current_term;}} }
  
  inline void remove_from_invariant (invariant_term* it) {
    invariant_term *next_it;
#ifdef tracentry
    trace_rfi += 1;
#endif
#ifdef debug
    if (it->invariant_term_magic_no != invariant_term_magic)
      die ("attempting to remove a non-invariant");
#endif
    int power, svi;
    for (int i = 0; i < number_to_simplify; i++) {
      svi = pass_fns[i + pass_fn_index];
#ifdef debug
      if ((svi < 0) || (svi > number_of_functions))
	die ("bad sort variable index in inv:rem_from_inv");
#endif
      if (sort_variable[svi] >= 0) {
	power = abs (it->factor_powers[sort_variable[svi]]);
#ifdef debug
	if (power >= simplify_size)
	  die ("remove - invariant power exceeds simplify_size");
#endif
	if (invariant_head[i * simplify_size + power] == it) {    // it's on the front of the list
	  next_it = it->next_invariant_term (i);
#ifdef debug
	  if (next_it != NULL)
	    if (next_it->invariant_term_magic_no != invariant_term_magic)
	      die ("new head is a non-invariant");
#endif
	  invariant_head[i * simplify_size + power] = next_it;}}}
    it->remove_invariant_links (); }
};

//------------------------------------------------------------------------
// class component - element representing a tensor component
//------------------------------------------------------------------------

class component : public element {

 public:

#ifdef debug
  int component_magic_no;
#endif

  long long indices;                 // stored as up to 10 6-bit indexes, early index is low order
                                     //     index value of 63 indicates undefined value (to be summed)
  sum* value;

  component (hash*, int*);
  component (hash*, int*, sum*);
  void send (term*);

  inline ~component () {
#ifdef tracentry
    trace_kco += 1;
#endif
#ifdef debug
    component_magic_no = -component_magic;
#endif
    if (!value->zeroq ())
      die ("attempting to delete nonzero component");
    indices = 0;
    delete value;
    value = NULL;}

  inline component* next_component () {
    component* next_one = (component*) next_element ();
#ifdef tracentry
    trace_nco += 1;
#endif
#ifdef debug
    if ((next_one != NULL) && (next_one->component_magic_no != component_magic))
      die ("next link is not a component");
#endif
    return next_one; }
};

//------------------------------------------------------------------------
// global variables - those which depend on class definitions above
//------------------------------------------------------------------------

extern term* function_terms[];   // a list of terms - NOT stored as a hash
                                 // list is normalized so coefficient of 1st term is 1
extern sum* functions[];         // functions represented as sums
extern sum* derivatives[];       // derivatives of each function wrt each coord

extern sum* scacurv;

extern int factor_powers_temp[];
extern int factor_powers_temp_d[];
extern int factor_powers_temp_s[];
extern int factor_powers_temp_i[];
extern int factor_powers_dim[];
extern int factor_powers_dimm2[];
extern int factor_powers_dimdimm1[];
extern int gcd_factor_powers[];
