//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// ntuple.h       

//------------------------------------------------------------------------
// global variables - those which do not depend on class definitions
//------------------------------------------------------------------------

extern int epsilon_length;
extern int* which_epsilon;
extern int number_of_epsilons;

//------------------------------------------------------------------------
// class ntuple - hash object whose elements are tensor components
//------------------------------------------------------------------------

class ntuple : public hash {

 protected:

  component* current_component;

  long long current_free_index;        // current free index list index (long long for parallize compatibility)
  int free_index_length;
  long long* free_index_list;          // list of free component indices (matrix represented as 
                                       //                                 an array of long longs)
  int* index_list;                     // current set of indices

  long long* indices_list;             // used in contraction

  long long current_sum_index;         // current sum list index (long long for parallize compatibility)
  int sum_list_length;
  component** sum_list;                // list of components for contractions

 public:

  bool metric;                         // true for g and ginv 
                                       // (used to keep indices in order when raising and lowering)
  bool* upper;                         // index flags
  bool* antisymmetric;                 // designates an (anti)symmetric index (wrt a neighbor index)
  bool* symmetric;                     //    only come in pairs
  bool* spinor;

  component** comparray;               // array of dim^rank pointers to components
  int casize;                          // size of comparray
  
#ifdef debug
  int ntuple_magic_no;
#endif

  ntuple ();
  ntuple (int);
  ntuple (int, bool*, bool*, bool*, bool*);
  ~ntuple ();
  void ntuple_init ();
  bool make_free_index_list (bool);
  int make_sum_list (int, ntuple**, bool);
  bool parallelize (sum*);
  bool parallelize (ntuple*);
  bool free_parallelize (ntuple*);
  bool parallelize_dual (sum*);
  int contract (int, ntuple**, ntuple*, bool);
  int contract (int, ntuple**, sum*, bool);
  ntuple* raise (int, int*);            // contracts on g inverse
  ntuple* lower (int, int*);            // contracts on g
  ntuple* covariant_derivative ();      // computes covariant derivative  (lower indices!)
  void zero ();
  void copy (ntuple*);
  ntuple* clone ();
  void add (ntuple*);
  void fill ();
  void simplify ();
                                       // contraction
  ntuple* self_contract (int, int);
  ntuple* ntuple_contract (int, ntuple**, int*);
  ntuple* dual (int, int*);
  ntuple* gamma (int, int*);

  sum* scalar ();
  sum* scalar_invariant (int, ntuple**, int*, bool);
  sum* dual_invariant (int, ntuple**, int*, bool);

  void display (char*);                 // displays a component
  void display_add (char*);             // displays a component with +=
#ifdef debug
  void dump (char*);                    // dumps the n-tuple
#endif

  void tensor_build (int, int*, int*, int*, int*, long long*);

  inline void insert_component (component* component_param) {
#ifdef tracentry
    trace_ic += 1;
#endif
#ifdef debug
    if (component_param->component_magic_no != component_magic)
      die ("attempting to insert a non-component");
#endif
    insert_at_head ((element*) component_param); }

  inline void remove_component (component* component_param) {
#ifdef tracentry
    trace_rc += 1;
#endif
#ifdef debug
    if (component_param->component_magic_no != component_magic)
      die ("attempting to remove a non-component");
#endif
    remove_from_list ((element*) component_param);
    number_of_components -= 1;
    delete component_param;}

  inline component* first_component (int index) {
#ifdef tracentry
    trace_fc += 1;
#endif
#ifdef debug
    if ((head[index] != NULL) && (((component*) head[index])->component_magic_no != component_magic))
      die ("head is a non-component");
#endif
    return (component*) head[index]; }

  inline int free_index (int index) {
#ifdef tracentry
    trace_fi += 1;
#endif
#ifdef debug
    if ((current_free_index < 0) || (current_free_index >= free_index_length))
      die ("current_free_index out of bounds");
#endif
    int index_value;
    index_value = (free_index_list[current_free_index] >> (6 * index)) & 63;
    if (index_value == 63)
      index_value = 0x80000001;
    return index_value;}

  inline long long ca_index (int* indices_param) {           // returns index into comparray for this component
#ifdef debug
    if (casize <= 0)
      die ("comparray not initialized");
#endif
    long long nzind = indices_param[0];
    for (int i = 1; i < rank; i++)
      if (spinor[i - 1])
	nzind = nzind * spinor_dimension + indices_param[i];
      else
	nzind = nzind * dimension + indices_param[i];
#ifdef debug
    if (nzind >= casize)
      die ("nzind out of bounds");
#endif
    return nzind; }

  inline component* sub (int* indices_param) {          // component selection
#ifdef tracentry
    trace_csu += 1;
#endif
    component* current_component;
    last_hash = -1;
    if (rank <= 6)
      return comparray[ca_index (indices_param)];
    hash_code (indices_param);
    current_component = first_component (last_hash);
    long long inds = 0;
    for (int i = 0; i < rank; i++)
      inds |= (((long long) indices_param[i]) & 63) << (6 * i);
    while (current_component != NULL) {
#ifdef debug
      if (current_component->component_magic_no != component_magic)
	die ("found a non-component");
#endif
      if (current_component->indices == inds)
	return current_component;
      current_component = current_component->next_component ();}
    return NULL;}

  inline void ntuple::set (int* indices, term* new_term) {
#ifdef tracentry
    trace_nst += 1;
#endif
#ifdef debug
    if (new_term == NULL)
      die ("attempting to set a component to null term");
    if (new_term->term_magic_no != term_magic)
      die ("attempting to set a non-term");
#endif
    component* current_component = sub (indices);
    if (current_component == NULL) {
      current_component = new component ((hash*) this, indices);
      if (current_component == NULL)
	die ("allocating component");
      hash_code (indices);
      insert_component (current_component);}
    else {
      if (current_component->value == NULL) {
	current_component->value = new sum;
	if (current_component->value == NULL)
	  die ("allocating new value for component");}
      else {
#ifdef debug
	die ("replacing a nonzero sum with a term");
#endif
	(current_component->value)->zero ();}}
    if (rank <= 6)
      comparray[ca_index (indices)] = current_component;
    (current_component->value)->hash_code (new_term->factor_powers);
    (current_component->value)->insert_term (new_term); }
  
  inline void ntuple::set (int* indices, sum* new_sum) {
#ifdef tracentry
    trace_nss += 1;
#endif
#ifdef debug
    if (new_sum == NULL)
      die ("attempting to set a component to null sum");
    if (new_sum->sum_magic_no != sum_magic)
      die ("attempting to set a non-sum");
#endif
    bool zsum = new_sum->zeroq ();
    component* current_component = sub (indices);
    if (current_component == NULL) {
      if (zsum)
	return;
      current_component = new component ((hash*) this, indices, new_sum);
      if (current_component == NULL)
	die ("allocating component");
      hash_code (indices);
      insert_component (current_component);}
    else {
      if (current_component->value != NULL) {
	(current_component->value)->zero ();
#ifdef debug
	die ("replacing a nonzero sum with a sum");
#endif
	delete current_component->value;}
      if (zsum) {
	if (rank <= 6) {
	  hash_code (indices);
	  comparray[ca_index (indices)] = NULL;}
	remove_component (current_component);
	return;}}
    if (rank <= 6)
      comparray[ca_index (indices)] = current_component;
    current_component->value = new_sum; }

  inline void scalar_multiply (long long factor) {
#ifdef tracentry
    trace_nsmu += 1;
#endif
    component *current_component;
    if (factor == 1)
      return;
    for (int k = 0; k < regular_hash_size; k++) {
      current_component = first_component (k);
      while (current_component != NULL) {
	if (current_component->value != NULL)
	  (current_component->value)->scalar_multiply (factor);
	current_component = current_component->next_component ();}} }

  inline void scalar_multiply (int *fraction_powers) {
#ifdef tracentry
    trace_nsmu += 1;
#endif
    component *current_component;
    for (int k = 0; k < regular_hash_size; k++) {
      current_component = first_component (k);
      while (current_component != NULL) {
	if (current_component->value != NULL)
	  (current_component->value)->scalar_multiply (fraction_powers);
	current_component = current_component->next_component ();}} }

  inline ntuple* multiply (sum* su) {
#ifdef tracentry
    trace_nms += 1;
#endif
    int inds[rank];
    long long one = 1;
    sum *newsum, *slist[2];
    ntuple* newnt = new ntuple (rank, upper, antisymmetric, symmetric, spinor);
    component *current_component;
    if (su->zeroq ())
      return newnt;
    slist[0] = su;
    for (int k = 0; k < regular_hash_size; k++) {
      current_component = first_component (k);
      while (current_component != NULL) {
	if (current_component->value != NULL) {
	  slist[1] = current_component->value;
	  newsum = new sum;
	  newsum->multiply (2, slist, &one);
	  for (int i = 0; i < rank; i++)
	    inds[i] = (current_component->indices >> (6 * i)) & 63;
	  newnt->set (&(inds[0]), newsum);}
	current_component = current_component->next_component ();}}
    return newnt; }
  
  inline void ntuple::add (int* indices, sum* new_sum) {
#ifdef tracentry
    trace_na += 1;
#endif
#ifdef debug
    if (new_sum == NULL)
      die ("attempting to add a null sum to a component");
    if (new_sum->sum_magic_no != sum_magic)
      die ("attempting to add a non-sum");
#endif
    if (new_sum->zeroq ())
      return;
    component* current_component = sub (indices);
    if (current_component == NULL) {
      current_component = new component ((hash*) this, indices);
      if (current_component == NULL)
	die ("allocating component");
      hash_code (indices);
      insert_component (current_component);}
    else {
      if (current_component->value == NULL) {
	current_component->value = new sum;
	if (current_component->value == NULL)
	  die ("allocating new value for component");}}
    sum* multiply_list = new_sum;
    long long one = 1;
    (current_component->value)->multiply (1, &multiply_list, &one);
    if ((current_component->value)->zeroq ()) {
      if (rank <= 6) {
	hash_code (indices);
	comparray[ca_index (indices)] = NULL;}
      remove_component (current_component);}
    else {
      if (rank <= 6)
	comparray[ca_index (indices)] = current_component;} }

  inline void ntuple::fast_lorentz_raise (int ind) {
#ifdef tracentry
    trace_flr += 1;
#endif
#ifdef debug
    if ((ind < 0) || (ind >= rank))
      die ("invalid index in fast_lorentz_raise");
#endif
    component *current_component;
    for (int k = 0; k < regular_hash_size; k++) {
      current_component = first_component (k);
      while (current_component != NULL) {
	if (((current_component->indices >> (6 * ind)) & 63) == 0)
	  if (current_component->value != NULL)
	    (current_component->value)->scalar_multiply (-1);
	current_component = current_component->next_component ();}}
    upper[ind] = true; }

  inline void ntuple::transpose () {        // transposes first two indices
#ifdef tracentry
    trace_ntr += 1;
#endif
    int ind, ind1;
    int inds0 [rank], inds1[rank];
    component *current_component, *tempc;
    bool temp;
    temp = upper[0];
    upper[0] = upper[1];
    upper[1] = temp;
    temp = spinor[0];
    spinor[0] = spinor[1];
    spinor[1] = temp;
    for (int k = 0; k < regular_hash_size; k++) {
      current_component = first_component (k);
      while (current_component != NULL) {
	if (rank <= 6)
	  for (int i = 0; i < rank; i++)
	    inds0[i] = (current_component->indices >> (6 * i)) & 63;
	ind = current_component->indices & (63*64 + 63);
	current_component->indices = current_component->indices - ind;
	ind = ((ind & (63*64)) >> 6) + ((ind & 63) << 6);
	current_component->indices = current_component->indices + ind;
	if (rank <= 6) {
	  for (int i = 0; i < rank; i++)
	    inds1[i] = (current_component->indices >> (6 * i)) & 63;
	  if (inds0[0] < inds0[1]) {
	    ind = ca_index (inds0);
	    tempc = comparray[ind];
	    ind1 = ca_index (inds1);
	    comparray[ind] = comparray[ind1];
	    comparray[ind1] = tempc;}}
	current_component = current_component->next_component ();}} }
};

//------------------------------------------------------------------------
// class connection - Christoffel or spin connection
//------------------------------------------------------------------------

class connection : public ntuple {

 public:

#ifdef debug
  int connection_magic_no;
#endif

  connection ();
};

//------------------------------------------------------------------------
// class curvature - Riemann curvature
//------------------------------------------------------------------------

class curvature : public ntuple {

 public:

#ifdef debug
  int curvature_magic_no;
#endif

  curvature ();
};  

//------------------------------------------------------------------------
// global variables - those which depend on class definitions above
//------------------------------------------------------------------------

extern ntuple* g;                    // spacetime metric
extern ntuple* g_inverse;
extern sum* inverse_sqrt_det_g;
extern connection* christoffel;      // first index is upper
extern curvature* riemann;           // 1st index is upper
extern ntuple* riemann_all_low;
extern ntuple* ricci;
extern ntuple* traceless_ricci;
extern ntuple* einstein_curv;
extern ntuple* g_curv;
extern ntuple* weyl;
